// Software Version 8.517 (y.mdd)
#include <p18f4320.h>
//-----------------------
// Data size definition -
//-----------------------
#define BYTE unsigned char
#define WORD unsigned int
#define LONG unsigned long

//------------------------
//- Functions prototypes -
//------------------------
void InterruptHandler(void);

//------------------
//- I/O definition -
//------------------
#define ADDR	0x0388
#define DATA	ADDR+1
#define STATUS	ADDR

#define ISA_RESET   LATBbits.LATB5  // output RESET
#define ISA_IOR     LATEbits.LATE0  // output IOR
#define ISA_IOW     LATEbits.LATE1  // output IOR
#define ISA_IRQ5    PORTBbits.RB0   // input IRQ5

//------------------------
//- Variables definition -
//------------------------
#pragma udata gpr1 // place in extended area (0x100-0x1FF)
static BYTE _area[64], _bumidi[8], _buvel[8], _tabv1[8], _tabv2[8];

#pragma udata // place in normal area
BYTE immb0, servpt, inpupt;
char *area = &_area[0];
char *bumidi = &_bumidi[0];
char *buvel = &_buvel[0];
char *tabv1 = &_tabv1[0];
char *tabv2 = &_tabv2[0];
BYTE bufpre[9], canale[8], triger[8], address[8];
BYTE polvol[8], prevel[8], hreg[8], lreg[8];
BYTE polybf, comodo, peadd, fltuda, cpres, note;
BYTE wait7, ddsalt, midchn, comm9e, contmd;
BYTE total, delta, ksr, hchan;
BYTE pusha;
BYTE excom;

//------------------------
//- Constants definition -
//------------------------
#include "tables.c"

//--------------------
//- Interrupt vector -
//--------------------
#pragma code InterruptVector = 0x08
void InterruptVector(void)
{
_asm
    goto InterruptHandler
_endasm
}

//---------------------
//- Interrupt handler -
//---------------------
#pragma code
#pragma interrupt InterruptHandler
void InterruptHandler(void)
{
    if(INTCONbits.TMR0IF) // if Tmr0 Overflow interrupt
        {
        INTCONbits.TMR0IF = 0; // clear int flag
        }
    if(PIR1bits.RCIF) // if serial port RX interrupt
        {
        area[inpupt++] = RCREG; // store serial data to queue
        inpupt &= 0x3F; // circular buffer range is 0..0x3F
        }
}

void OutCom(BYTE d)
{
    while(TXSTAbits.TRMT == 0); // wait for transmit register empty (TRMT=1)
    TXREG = d; // transfer data to transmit register
}

//---------------------
//- ISA BUS routines  -
//---------------------
BYTE inportb(WORD addr)
{
BYTE data;

    LATC = addr & 0x3F; // set address (ISA_SA5..SA0)
    ISA_IOR = 0; // activate IO READ signal
    data = PORTD; // get data (ISA_SD7..SD0)
    ISA_IOR = 1; // de-activate IO READ signal
    return data; 
}

void outportb(WORD addr, BYTE data)
{
    LATC = addr & 0x3F; // set address (ISA_SA5..SA0)
    LATD = data; // prepare data (ISA_SD7..SD0)
    TRISD = 0x00; // set data bus as output
    ISA_IOW = 0; // activate IO WRITE signal
    ISA_IOW = 1; // de-activate IO WRITE signal
    TRISD = 0xFF; // set data bus as input
}

void WrReg(BYTE ix, BYTE dt)
{
BYTE c;

    outportb(ADDR, ix);
    for(c = 0; c < 3; ++c) // 3.3 uS after writing ADDR
        Nop();
    outportb(DATA, dt);
    for(c = 0; c < 23; ++c) // 23 uS after writing DATA
        Nop();
}

//-----------------
//- MAIN PROGRAM  -
//-----------------

#include "vmain2.c" // main (complete version)

void InitFM(void)
{
BYTE a, y;

    ClrWdt(); // kick the watchdog
    for(y = 0;;)
        {
        if((a = init_table[y++]) == 0)
            break;
        WrReg(a, init_table[y++]);
        }
}

void InitPresets(BYTE instr)
{
BYTE x;

	for(x = 0; x < 8; ++x) // for all the channels
		LoadPreset(instr, x); // set the instrument
}

void MidiParser(void)
{
BYTE a, x, y;

	if(servpt == inpupt) // if no MIDI data pending
		return; // do nothing

    y = area[servpt++]; // get data from serial queue
    servpt &= 0x3F; // circular buffer range is 0..0x3F
	if(y > 0x7F) // if MIDI command
		{
        wait7 = 0; // reset flag for pitch wheel change
		switch(a = y & 0xf0) // received byte (high nibble only)
			{
			case 0x80: // note OFF
			case 0x90: // note ON
				comm9e = a;
				if((y & 0x0f) != midchn) // if not my channel
					{
                    // attenzione: potrebbero esserci altri bytes dopo i primi 2
					ddsalt = 2; // ignore next 2 bytes
					return;
					}
				contmd = 1; // set flag for midi control
				return;
			case 0xc0: // Program change
				if((y & 0x0f) != midchn) // if not my channel
					{
					ddsalt = 1; // ignore next byte
					return;
					}
				cpres = 1; // set flag for program change
				return;
			case 0xd0: // channel pressure
				ddsalt = 1; // ignore next byte
				return;
			case 0xa0: // polyphonic key pressure
            case 0xb0: // channel mode message
                contmd = 0; // reset flag for midi control
        		cpres = 0; // reset flag for program change
				ddsalt = 2; // ignore next 2 bytes
				return;
			case 0xe0: // pitch wheel change
                contmd = 0; // reset flag for midi control
        		cpres = 0; // reset flag for program change
                wait7 = 1; // set flag for pitch wheel change
				return;
			}
		switch(y) // full received byte
			{
			case 0xf0: // start of system exclusive
				fltuda = 1; // set flag for system exclusive
				return;
			case 0xf2: // song position pointer
				ddsalt = 2; // ignore next 2 bytes
				return;
			case 0xf3: // song select
				ddsalt = 1; // ignore next byte
				return;
			case 0xf7: // end of system exclusive
				fltuda = 0; // reset flag for system exclusive
				return;
			default: // ignore F1,F4,F5,F6,F8..FF
				return;
			}
		}
    
    // if MIDI data
	if(fltuda != 0 || wait7 != 0) // if system exclusive message incoming or pitch wheel change
		return; // do nothing (ignore system exclusive)

	if(cpres != 0) // if program change flag set
		{
        InitPresets(y & 0x0f); // preset only in range 0..0x0F (1..16)
    	for(x = 0; x < 8; ++x) // clear all midi buffer
    		bumidi[x] = 0;
		cpres = 0; // reset flag for program change
		return;
		}

	if(ddsalt != 0) // if data to be ignored
		{
		--ddsalt; // decrement byte counter
		return;
		}

	if(contmd != 0) // if flag for midi control set
		{
		contmd = 0; // reset flag for midi control
        note = y; // store first byte after command (0x80 or 0x90)
		return;
		}

	contmd = 1; // set flag as 0x80 or 0x90 received (next notes)
	if(y == 0 || comm9e == 0x80) // if note OFF (NOTE ON with speed=0 or NOTE OFF)
		{
		for(x = 0; x < 8; ++x) // scan the midi input buffer
			{
			if((a = bumidi[x] & 0x7f) == note) // if note present (ON or OFF)
				{
				bumidi[x] = a; // turn off the NOTE ON flag
				break; // terminate loop
				}
			}
		return;
		}

    // if NOTE ON
	for(x = 0; x < 8; ++x) // scan the midi input buffer
		{
		if(note == (bumidi[x] & 0x7F)) // if note present in buffer, but NOTE OFF
			{
			bumidi[x] |= 0x80; // set the NOTE ON
        	buvel[x] = y; // store speed
			return;
			}
		}

    // if NOTE ON, not present in buffer
	for(x = 0; x < 8; ++x) // scan the midi input buffer for free space
		{
		if(bumidi[x] < 0x80) // if channel free (note off)
            {
        	bumidi[x] = note | 0x80; // store the note with NOTE ON flag
        	buvel[x] = y; // store speed
			return;
            }
		}
    // if buffer full, note discarded
}

void InitRam(void)
{
BYTE x;

	for(x = 0; x < 8; x++)
		{
		address[x] = x;
		bufpre[x] = canale[x] = triger[x] = 0;
		polvol[x] = 0;
		prevel[x] = hreg[x] = lreg[x] = 0;
		tabv1[x] = tabv2[x] = 0;
		bumidi[x] = buvel[x] = 0;
		}
	midchn = 0;
    fltuda = 0;
    cpres = 0;
    ddsalt = 0;
}

void main(void)
{
    ADCON1 = 0x0F;  // no analog inputs used
    TRISA = 0x80;   // 1000.0000 Port direction, 0=out
                    // .0 = 
                    // .1 = 
                    // .2 = 
                    // .3 = 
                    // .4 = 
                    // .5 =  
                    // .6 = OscOut
                    // .7 = OscIn
    LATA = 0xFF;    // 1111.1111 Port data
   
    TRISB = 0xC1;   // 1100.0001 Port direction, 0=out
                    // .0 = ISA_IRQ5
                    // .1 = 
                    // .2 = 
                    // .3 = 
                    // .4 = 
                    // .5 = ISA_RESET
                    // .6 = PGC
                    // .7 = PGD
    LATB = 0x21;    // 0010.0001 Port data
    
    TRISC = 0x80;   // 1000.0000 Port direction, 0=out
                    // .0 = ISA_SA0
                    // .1 = ISA_SA1
                    // .2 = ISA_SA2
                    // .3 = ISA_SA3
                    // .4 = ISA_SA4
                    // .5 = ISA_SA5
                    // .6 = TX
                    // .7 = RX
    LATC = 0xC0;    // 1100.0000 Port data
        
    TRISD = 0xFF;   // 1111.1111 Port direction, 0=out
                    // .0 = ISA_SD0
                    // .1 = ISA_SD1
                    // .2 = ISA_SD2
                    // .3 = ISA_SD3
                    // .4 = ISA_SD4
                    // .5 = ISA_SD5
                    // .6 = ISA_SD6
                    // .7 = ISA_SD7
    LATD = 0xFF;    // 1111.1111 Port data 
    
    TRISE = 0x08;   // 0000.1000 Port direction, 0=out (no PSP mode)
                    // .0 = ISA_IOR
                    // .1 = ISA_IOW
                    // .2 = 
                    // .3 = MCLR/VPP
    LATE = 0x0B;    // 0000.1011 Port data
        
    OSCCON = 0x7F;  // 0111.1111 Set oscillator control register

    SPBRG = 3; // set 31250 bps @ 8MHz   
    TXSTA = 0xA3; // Enable serial TX
    RCSTA = 0x91; // Enable serial port1, enable serial RX
    PIE1bits.RCIE = 1; // enable Serial RX interrupt
    PIR1bits.RCIF = 0; // clear any pending interrupt flag

    T0CON = 0xD3; // 1101.0011 Enable TMR0 as 8 bit, presc=16
    INTCONbits.TMR0IF = 0; // clear any pending interrupt flag
    INTCONbits.TMR0IE = 1; // Enable TMR0 Overflow interrupt (1.6384mS @10MHz)

    INTCONbits.PEIE = 1; // peripheral interrupts enable
    INTCONbits.GIE = 1; // Global interrupt enable

    ISA_RESET = 0; // remove the ISA reset condition
    InitRam(); // ram variables initialization
    InitFM(); // init the FM registers
    InitPresets(0); // load default preset for all channels

	for(;;)
		{
        ClrWdt(); // kick the watchdog
		MidiParser(); // midi input data parser
        Upconv(); // convert midi in to internal format
		Poly(); // polyphonic handler
		}
}
